@@ -62,6 +62,39 @@ def test_wrapcontents():
     page.clean_contents()
 
 
+def test_page_clean_contents():
+    """Assert that page contents cleaning actually is invoked."""
+    doc = fitz.open()
+    page = doc.new_page()
+
+    # draw two rectangles - will lead to two /Contents objects
+    page.draw_rect((10, 10, 20, 20))
+    page.draw_rect((20, 20, 30, 30))
+    assert len(page.get_contents()) == 2
+    assert page.read_contents().startswith(b"q") == False
+
+    # clean / consolidate into one /Contents object
+    page.clean_contents()
+    assert len(page.get_contents()) == 1
+    assert page.read_contents().startswith(b"q") == True
+
+
+def test_annot_clean_contents():
+    """Assert that annot contents cleaning actually is invoked."""
+    doc = fitz.open()
+    page = doc.new_page()
+    annot = page.add_highlight_annot((10, 10, 20, 20))
+
+    # the annotation appearance will not start with command b"q"
+    assert annot._getAP().startswith(b"q") == False
+
+    # invoke appearance stream cleaning and reformatting
+    annot.clean_contents()
+
+    # appearance stream should now indeed start with command b"q"
+    assert annot._getAP().startswith(b"q") == True
+
+
 def test_config():
     assert fitz.TOOLS.fitz_config["py-memory"] in (True, False)
 
@@ -212,7 +245,7 @@ def test_texttrace():
         tt = page.get_texttrace()
     t = time.time() - t
     print( f'test_texttrace(): t={t!r}')
-    
+
     # Repeat, this time writing data to file.
     import json
     path = f'{scriptdir}/resources/test_texttrace.txt'
@@ -222,6 +255,57 @@ def test_texttrace():
             tt = page.get_texttrace()
             print( f'page {i} json:\n{json.dumps(tt, indent="    ")}', file=f)
 
+
+def test_2533():
+    """Assert correct char bbox in page.get_texttrace().
+
+    Search for a unique char on page and confirm that page.get_texttrace()
+    returns the same bbox as the search method.
+    """
+    if hasattr(fitz, 'mupdf') and not fitz.g_use_extra:
+        print('Not running test_2533() because rebased with use_extra=0 known to fail')
+        return
+    doc = fitz.open(os.path.join(scriptdir, "resources", "test_2533.pdf"))
+    page = doc[0]
+    NEEDLE = "民"
+    ord_NEEDLE = ord(NEEDLE)
+    for span in page.get_texttrace():
+        for char in span["chars"]:
+            if char[0] == ord_NEEDLE:
+                bbox = fitz.Rect(char[3])
+                break
+    assert page.search_for(NEEDLE)[0] == bbox
+
+
+def test_2506():
+    """Ensure expected font size across text writing angles."""
+    doc = fitz.open()
+    page = doc.new_page()
+    point = fitz.Point(100, 300)  # insertion point
+    fontsize = 11  # fontsize
+    text = "Hello"  # text
+    angles = (0, 30, 60, 90, 120)  # some angles
+
+    # write text with different angles
+    for angle in angles:
+        page.insert_text(
+            point, text, fontsize=fontsize, morph=(point, fitz.Matrix(angle))
+        )
+
+    # ensure correct fontsize for get_texttrace() - forgiving rounding problems
+    for span in page.get_texttrace():
+        print(span["dir"])
+        assert round(span["size"]) == fontsize
+
+    # ensure correct fontsize for get_text() - forgiving rounding problems
+    for block in page.get_text("dict")["blocks"]:
+        for line in block["lines"]:
+            print(line["dir"])
+            for span in line["spans"]:
+                print(span["size"])
+                assert round(span["size"]) == fontsize
+
+
 def test_2108():
     doc = fitz.open(f'{scriptdir}/resources/test_2108.pdf')
     page = doc[0]
@@ -230,9 +314,9 @@ def test_2108():
     page.add_redact_annot(rect)
     page.apply_redactions()
     text = page.get_text()
-    
+
     text_expected = b'Frau\nClaire Dunphy\nTeststra\xc3\x9fe 5\n12345 Stadt\nVertragsnummer:  12345\nSehr geehrte Frau Dunphy,\nText\nMit freundlichen Gr\xc3\xbc\xc3\x9fen\nTestfirma\nVertrag:\n  12345\nAnsprechpartner:\nJay Pritchet\nTelefon:\n123456\nE-Mail:\ntest@test.de\nDatum:\n07.12.2022\n'.decode('utf8')
-    
+
     if 1:
         # Verbose info.
         print(f'test_2108(): text is:\n{text}')
@@ -257,7 +341,7 @@ def test_2108():
         print(f'ok3={ok3}')
 
         print(f'')
-    
+
     print(f'fitz.mupdf_version_tuple={fitz.mupdf_version_tuple}')
     if fitz.mupdf_version_tuple >= (1, 21, 2):
         print('Asserting text==text_expected')
@@ -360,8 +444,8 @@ def test_2182():
     for annot in page.annots():
         print(annot)
     print(f'test_2182() finished')
-    
-  
+
+
 def test_2246():
     """
     Test / confirm identical text positions generated by
@@ -421,3 +505,22 @@ def test_2246():
     assert bbox_count(180) == 1
     assert bbox_count(270) == 1
 
+
+def test_2430():
+    """Confirm that multiple font property checks will not destroy Py_None."""
+    font = fitz.Font("helv")
+    for i in range(1000):
+        _ = font.flags
+
+def test_2692():
+    document = fitz.Document(f'{scriptdir}/resources/2.pdf')
+    for page in document:
+        pix = page.get_pixmap(clip=fitz.Rect(0,0,10,10))
+        dl = page.get_displaylist(annots=True)
+        pix = dl.get_pixmap(
+                matrix=fitz.Identity,
+                colorspace=fitz.csRGB,
+                alpha=False,
+                clip=fitz.Rect(0,0,10,10),
+                )
+    
